"""DROID circuits. These circuits are auto-generated from circuits.json."""

from dataclasses import dataclass, field
from typing import Optional

from metapatch.circuits.base import DroidCircuit
from metapatch.circuits import circuit_types as ctype


__droid_version__ = "magenta-3"


@dataclass
class Droid(DroidCircuit):
    """General DROID controls.

    Args:
        ledbrightness (cv):
          Let's you dim all of the 24 LEDs of the master and the G8. This is mainly for
          those who think they are too bright. But since this parameter can be CV-
          controlled, you could of course also do funny things with it. Beware: if you
          set this to zero, the LEDs will be completely dark. This also includes
          possible error messages.
        maxslope1 (cv):
          Sets a threshold for a voltage change between two samples until the internal
          logic of the outputs assumes that this step is intentional and should not be
          smoothed out. A typical case where you do not want smoothing is the pitch
          output of a sequencer.  The default value is 0.25. A value of 0.0 turns off
          smoothing altogether since the slightest voltage change is considered an
          intentional jump.
        maxslope2 (cv):
          Sets a threshold for a voltage change between two samples until the internal
          logic of the outputs assumes that this step is intentional and should not be
          smoothed out. A typical case where you do not want smoothing is the pitch
          output of a sequencer.  The default value is 0.25. A value of 0.0 turns off
          smoothing altogether since the slightest voltage change is considered an
          intentional jump.
        maxslope3 (cv):
          Sets a threshold for a voltage change between two samples until the internal
          logic of the outputs assumes that this step is intentional and should not be
          smoothed out. A typical case where you do not want smoothing is the pitch
          output of a sequencer.  The default value is 0.25. A value of 0.0 turns off
          smoothing altogether since the slightest voltage change is considered an
          intentional jump.
        maxslope4 (cv):
          Sets a threshold for a voltage change between two samples until the internal
          logic of the outputs assumes that this step is intentional and should not be
          smoothed out. A typical case where you do not want smoothing is the pitch
          output of a sequencer.  The default value is 0.25. A value of 0.0 turns off
          smoothing altogether since the slightest voltage change is considered an
          intentional jump.
        maxslope5 (cv):
          Sets a threshold for a voltage change between two samples until the internal
          logic of the outputs assumes that this step is intentional and should not be
          smoothed out. A typical case where you do not want smoothing is the pitch
          output of a sequencer.  The default value is 0.25. A value of 0.0 turns off
          smoothing altogether since the slightest voltage change is considered an
          intentional jump.
        maxslope6 (cv):
          Sets a threshold for a voltage change between two samples until the internal
          logic of the outputs assumes that this step is intentional and should not be
          smoothed out. A typical case where you do not want smoothing is the pitch
          output of a sequencer.  The default value is 0.25. A value of 0.0 turns off
          smoothing altogether since the slightest voltage change is considered an
          intentional jump.
        maxslope7 (cv):
          Sets a threshold for a voltage change between two samples until the internal
          logic of the outputs assumes that this step is intentional and should not be
          smoothed out. A typical case where you do not want smoothing is the pitch
          output of a sequencer.  The default value is 0.25. A value of 0.0 turns off
          smoothing altogether since the slightest voltage change is considered an
          intentional jump.
        maxslope8 (cv):
          Sets a threshold for a voltage change between two samples until the internal
          logic of the outputs assumes that this step is intentional and should not be
          smoothed out. A typical case where you do not want smoothing is the pitch
          output of a sequencer.  The default value is 0.25. A value of 0.0 turns off
          smoothing altogether since the slightest voltage change is considered an
          intentional jump.
        lpfilter1 (fraction):
          Configures a digital low pass filter on the output in order to smooth out
          digital noise resulting from the 's main loop. This loop is running somewhere
          between 3 and 6 kHz – depending on the number of circuits you use.  Per
          default this filter is set to 0.25 – which means a mild filtering – thus still
          allowing fast and snappy envelopes and other rapidly changing signals while
          filtering away most of the digital artefacts.  If you use an output for a slow
          envelope that is combined with an audio path in a way that you hear digital
          artifacts then increase that value. This is e.g. the case if you modulate a
          VCA that in turn modulates a very low pitched audio wave with very few
          harmonics (such as a sine or triangle wave).  The maximum value of 1.0 leads
          to a very strong filtering – i.e. removing all fast transients. Snappy
          envelopes will be smoothed out heavily. Square wave LFOs will be converted
          into lower level almost sine waves.
        lpfilter2 (fraction):
          Configures a digital low pass filter on the output in order to smooth out
          digital noise resulting from the 's main loop. This loop is running somewhere
          between 3 and 6 kHz – depending on the number of circuits you use.  Per
          default this filter is set to 0.25 – which means a mild filtering – thus still
          allowing fast and snappy envelopes and other rapidly changing signals while
          filtering away most of the digital artefacts.  If you use an output for a slow
          envelope that is combined with an audio path in a way that you hear digital
          artifacts then increase that value. This is e.g. the case if you modulate a
          VCA that in turn modulates a very low pitched audio wave with very few
          harmonics (such as a sine or triangle wave).  The maximum value of 1.0 leads
          to a very strong filtering – i.e. removing all fast transients. Snappy
          envelopes will be smoothed out heavily. Square wave LFOs will be converted
          into lower level almost sine waves.
        lpfilter3 (fraction):
          Configures a digital low pass filter on the output in order to smooth out
          digital noise resulting from the 's main loop. This loop is running somewhere
          between 3 and 6 kHz – depending on the number of circuits you use.  Per
          default this filter is set to 0.25 – which means a mild filtering – thus still
          allowing fast and snappy envelopes and other rapidly changing signals while
          filtering away most of the digital artefacts.  If you use an output for a slow
          envelope that is combined with an audio path in a way that you hear digital
          artifacts then increase that value. This is e.g. the case if you modulate a
          VCA that in turn modulates a very low pitched audio wave with very few
          harmonics (such as a sine or triangle wave).  The maximum value of 1.0 leads
          to a very strong filtering – i.e. removing all fast transients. Snappy
          envelopes will be smoothed out heavily. Square wave LFOs will be converted
          into lower level almost sine waves.
        lpfilter4 (fraction):
          Configures a digital low pass filter on the output in order to smooth out
          digital noise resulting from the 's main loop. This loop is running somewhere
          between 3 and 6 kHz – depending on the number of circuits you use.  Per
          default this filter is set to 0.25 – which means a mild filtering – thus still
          allowing fast and snappy envelopes and other rapidly changing signals while
          filtering away most of the digital artefacts.  If you use an output for a slow
          envelope that is combined with an audio path in a way that you hear digital
          artifacts then increase that value. This is e.g. the case if you modulate a
          VCA that in turn modulates a very low pitched audio wave with very few
          harmonics (such as a sine or triangle wave).  The maximum value of 1.0 leads
          to a very strong filtering – i.e. removing all fast transients. Snappy
          envelopes will be smoothed out heavily. Square wave LFOs will be converted
          into lower level almost sine waves.
        lpfilter5 (fraction):
          Configures a digital low pass filter on the output in order to smooth out
          digital noise resulting from the 's main loop. This loop is running somewhere
          between 3 and 6 kHz – depending on the number of circuits you use.  Per
          default this filter is set to 0.25 – which means a mild filtering – thus still
          allowing fast and snappy envelopes and other rapidly changing signals while
          filtering away most of the digital artefacts.  If you use an output for a slow
          envelope that is combined with an audio path in a way that you hear digital
          artifacts then increase that value. This is e.g. the case if you modulate a
          VCA that in turn modulates a very low pitched audio wave with very few
          harmonics (such as a sine or triangle wave).  The maximum value of 1.0 leads
          to a very strong filtering – i.e. removing all fast transients. Snappy
          envelopes will be smoothed out heavily. Square wave LFOs will be converted
          into lower level almost sine waves.
        lpfilter6 (fraction):
          Configures a digital low pass filter on the output in order to smooth out
          digital noise resulting from the 's main loop. This loop is running somewhere
          between 3 and 6 kHz – depending on the number of circuits you use.  Per
          default this filter is set to 0.25 – which means a mild filtering – thus still
          allowing fast and snappy envelopes and other rapidly changing signals while
          filtering away most of the digital artefacts.  If you use an output for a slow
          envelope that is combined with an audio path in a way that you hear digital
          artifacts then increase that value. This is e.g. the case if you modulate a
          VCA that in turn modulates a very low pitched audio wave with very few
          harmonics (such as a sine or triangle wave).  The maximum value of 1.0 leads
          to a very strong filtering – i.e. removing all fast transients. Snappy
          envelopes will be smoothed out heavily. Square wave LFOs will be converted
          into lower level almost sine waves.
        lpfilter7 (fraction):
          Configures a digital low pass filter on the output in order to smooth out
          digital noise resulting from the 's main loop. This loop is running somewhere
          between 3 and 6 kHz – depending on the number of circuits you use.  Per
          default this filter is set to 0.25 – which means a mild filtering – thus still
          allowing fast and snappy envelopes and other rapidly changing signals while
          filtering away most of the digital artefacts.  If you use an output for a slow
          envelope that is combined with an audio path in a way that you hear digital
          artifacts then increase that value. This is e.g. the case if you modulate a
          VCA that in turn modulates a very low pitched audio wave with very few
          harmonics (such as a sine or triangle wave).  The maximum value of 1.0 leads
          to a very strong filtering – i.e. removing all fast transients. Snappy
          envelopes will be smoothed out heavily. Square wave LFOs will be converted
          into lower level almost sine waves.
        lpfilter8 (fraction):
          Configures a digital low pass filter on the output in order to smooth out
          digital noise resulting from the 's main loop. This loop is running somewhere
          between 3 and 6 kHz – depending on the number of circuits you use.  Per
          default this filter is set to 0.25 – which means a mild filtering – thus still
          allowing fast and snappy envelopes and other rapidly changing signals while
          filtering away most of the digital artefacts.  If you use an output for a slow
          envelope that is combined with an audio path in a way that you hear digital
          artifacts then increase that value. This is e.g. the case if you modulate a
          VCA that in turn modulates a very low pitched audio wave with very few
          harmonics (such as a sine or triangle wave).  The maximum value of 1.0 leads
          to a very strong filtering – i.e. removing all fast transients. Snappy
          envelopes will be smoothed out heavily. Square wave LFOs will be converted
          into lower level almost sine waves.
        m4faderspeed (fraction):
          Set the force / speed of the motor faders. Faster speeds need more electrical
          power and might wear off the faders faster. Too slow speeds might lead to poor
          operation. This value goes from 0.0 (slowest possible speed) to 1.0 (maximum
          speed). If you don't use this parameter, some reasonable default is used that
          depends on the firmware of the M4 module.
        statusdump (trigger):
          A trigger here does the same as a double “click” on the master's button: it
          creates a status dump file on the SD card. This trigger allows you automated
          control with a precise timing. Each dump needs a couple of milli seconds to
          write to the SD card. So better do not spam this input with a high frequency
          of triggers.
        m4notchpower (fraction):
          Set the force feedback power of the M4 motor fader units when they operate
          with virtual notches. The range is from 0 (minimum notch power) to 1 (maximum
          notch power). Note: 0 does not turn the notches off, there is still some
          minimal feedback. If you don't use this parameter, the notch force feedback
          operates at some default power, which is dependent on the M4 firmware version.
        calibrate (trigger):
          Immediately enter the calibration procedure, that's contained in the
          maintainance menu (only MASTER). Skips the menu. After calibration is done,
          resets.
        startcontrollerupgrade (trigger):
          Immediately starts the firmware upgrade procedure for controllers like M4 and
          E4. After one succesfull upgrade resets the master.
        startx7upgrade (trigger):
          Immediately starts the X7 firmware upgrade procedure (which is located at
          position 8 of the maintainance menu). After the upgrade of the X7 resets the
          master.
        clear (trigger):
          A trigger here sends a trigger to the clear input of all circuits that support
          this. That brings the state of those circuits to their start state. Circuits
          that have presets do keep those presets untouched. Just the current state is
          affected.  That trigger is not sent to circuits whose clear input is patched.
          Note: Just that part of the state is affected that is saved to the SD card.
          For example the algoquencer does not reset to the first step, it just clears
          it's current pattern.
        clearall (trigger):
          A trigger here sends a trigger to the clearall input of all circuits that
          support this. That's like a global factory reset for all of your circuits.
          Everything is set to its starting state, including all presets of those
          circuits.  That trigger is not sent to circuits whose clearall input is
          patched.  Note: Just that part of the state is affected that is saved to the
          SD card. For example the algoquencer does not reset to the first step, it just
          clears it's current pattern.
        comment: Add a comment in the droid ini file.

    """

    __ramsize__ = 72
    ledbrightness: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    maxslope1: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    maxslope2: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    maxslope3: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    maxslope4: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    maxslope5: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    maxslope6: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    maxslope7: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    maxslope8: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    lpfilter1: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    lpfilter2: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    lpfilter3: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    lpfilter4: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    lpfilter5: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    lpfilter6: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    lpfilter7: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    lpfilter8: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    m4faderspeed: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    statusdump: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    m4notchpower: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    calibrate: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    startcontrollerupgrade: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    startx7upgrade: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    clear: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    clearall: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )



@dataclass
class Outputcalibrator(DroidCircuit):
    """Tune the calibration of your CV outputs.

    Args:
        output (integer):
          Select the output to calibrate. This is a number from 1 to 8.
        referencepoint (integer):
          For each output, two voltages need to be adjusted: 0 V and 5 V. Select either
          0 for 0 V or 1 for 5 V.
        nudgeup (trigger):
          Increase the currently selected output voltage by one minimal step, to match
          the reference voltage.
        nudgedown (trigger):
          Decrease the currently selected output voltage by one minimal step, to match
          the reference voltage.
        save (trigger):
          A trigger here saves the changed calibration values to the internal flash of
          the master and the the SD card.
        cancel (trigger):
          A trigger here restores the previous calibration values from the internal
          flash.
        loaddefaults (trigger):
          A trigger here loads the default calibration values, which are not very
          precise, but a good starting point if you got totally lost.
        dirty (gate):
          Outputs 1 if the current calibration has been changed and needs to be saved.
        calibration (cv):
          Shows the difference between the current calibration of the selected output
          and reference voltage and its default calibration value.
        uncalibrated (fraction):
          Shows you the percentage of uncalibrated outputs. If all eight outputs are
          calibrated (differ from the default calibration value) this outputs 0.
        comment: Add a comment in the droid ini file.

    """

    __ramsize__ = 40
    output: Optional[str] = field(
            default=None,
            metadata=ctype.type_integer(ramsize=0)
    )
    referencepoint: Optional[str] = field(
            default=None,
            metadata=ctype.type_integer(ramsize=0)
    )
    nudgeup: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    nudgedown: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    save: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    cancel: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    loaddefaults: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    dirty: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    calibration: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    uncalibrated: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )






"""Formatting utilities."""

import textwrap

from typing import Optional

DEFAULT_LINEWIDTH = 88


def find_string_indent_level(text: str) -> int:
    """Find the indentation level of a string."""
    if not text.startswith(" "):
        return 0

    indent = 0
    for index in range(0, len(text)):
        if text[index] == " ":
            continue
        else:
            indent = index
            break

    if indent > 0:
        return int(indent / 4)

    return indent


def find_last(text: str, char: str, maxpos: Optional[int] = None) -> Optional[int]:
    """Find the last position of something in a string."""
    length = len(text)
    if not text.find(char):
        return None
    startpos = text.find(char) + 1
    pos = startpos
    while startpos < length:
        newpos = text.find(char, startpos, -1)
        if maxpos and newpos > maxpos:
            break
        if newpos > -1:
            pos = newpos
            startpos = newpos + 1
        else:
            break
    return pos


def format_line(text: str) -> str:
    """Format line so that long lines are nicely wrapped.

    Args:
        full_line: The entire line
        specific_item: The item that needs to be formatted
    """
    if len(text.rstrip()) < DEFAULT_LINEWIDTH:
        return text

    if '"' in text:
        # We break at the first assignment
        pos = text.find('"')
        startstr = text[:pos]
        if len(startstr) > DEFAULT_LINEWIDTH:
            # We should do something about that...
            startstr = format_line(startstr)

        # The text before might have become multiline, so we only count the final line
        start_last_line = startstr.splitlines()[-1]
        indent = find_string_indent_level(start_last_line)

        # We insert a (, go down a line, and then we have the current indentation as width.
        startstr = startstr.rstrip() + "(\n"

        # our new maximum string length is LINEWIDTH - indent - 4 (one level deeper than the last line)
        available_width = DEFAULT_LINEWIDTH - (4 * indent + 4)
        quoted = text[pos:]
        # We could split it at the original line ending, or we can just combine everything to one big string.
        lines = quoted.split("\\n")
        new_lines = []
        for index, line in enumerate(lines):
            if index == 0:
                start_indent = ""
            else:
                start_indent = "\\n"
            new_lines.extend(
                [
                    subline
                    for subline in textwrap.wrap(
                        line,
                        width=available_width,
                        initial_indent=start_indent,
                        subsequent_indent="\\n",
                    )
                ]
            )

        parent_indentstr = "    " * indent
        indentstr = "    " * (indent + 1)
        termination = '"'

        # Make groups of lines
        groups = []
        had_comma = False
        for index, part in enumerate(new_lines):
            if index == 0:
                groupstr = f"{indentstr}{part}"
            else:
                groupstr = f"{indentstr}{termination}{part}"

            if groupstr.endswith('",'):
                # This is the last one, we need to remove the comma.
                groupstr = groupstr[:-1]
                had_comma = True
            elif not groupstr.endswith(termination):
                groupstr += termination

            groups.append(groupstr)

        # Join everything up
        if had_comma:
            epilogue = "),\n"
        else:
            epilogue = ")\n"
        endstr = "\n".join(groups) + f"\n{parent_indentstr}{epilogue}"
        return startstr + endstr

    # If there's no quoted assignment, we split last possible parenthesis
    last_paren = find_last(text, "(", DEFAULT_LINEWIDTH)
    if last_paren:
        splitpos = last_paren + 1
        before = text[splitpos:]
        after = text[:splitpos]
        endsplit = find_last(after, ")")
        assert endsplit is not None
        content = after[:endsplit]
        inner = textwrap.indent(content, "    ")
        after_inner = after[endsplit:]
        return "\n".join([before, inner, after_inner])

    # If no parens, I don't know what to do.
    return text



"""DROID circuits. These circuits are auto-generated from circuits.json."""

from dataclasses import dataclass, field
from typing import Optional

from metapatch.circuits.base import DroidCircuit
from metapatch.circuits import circuit_types as ctype


__droid_version__ = "magenta-3"


@dataclass
class Fourstatebutton(DroidCircuit):
    """Button switching through 4 states (OBSOLETE).

    Args:
        button (trigger):
          The button.
        reset (trigger):
          A positive trigger here will reset the button to the first state.
        value1 (cv):
          The values that output should get when the four various states are active.
        value2 (cv):
          The values that output should get when the four various states are active.
        value3 (cv):
          The values that output should get when the four various states are active.
        value4 (cv):
          The values that output should get when the four various states are active.
        startvalue (integer):
          By setting this to 0, 1, 2 or 3 you set the initial state of the button when
          the is powered up to state 1, 2, 3 or 4. It also disabled the automatic saving
          of the button's state in the 's internal flash memory.
        output (cv):
          Depending on the current state of the button here the value of input1, input2,
          input3 or input4 will be copied.
        led (fraction):
          The LED in the button
        comment: Add a comment in the droid ini file.

    """

    __ramsize__ = 40
    button: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    reset: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    value1: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    value2: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    value3: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    value4: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    startvalue: Optional[str] = field(
            default=None,
            metadata=ctype.type_integer(ramsize=0)
    )
    output: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    led: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )



@dataclass
class Notchedpot(DroidCircuit):
    """Helper circuit for pots (OBSOLETE).

    Args:
        pot (fraction):
          Wire your pot here, e.g. P1.1
        notch (cv):
          Optionally set the notch size, if you do not like the default of 0.1. The
          maximum allowed value is 0.5. Greater values will be reduced to that.
        output (fraction):
          Your pot output comes here. It still goes from 0.0 to 1.0.
        bipolar (cv):
          Optional output with a range from -1.0 to 1.0, where the center notch is at
          0.0.
        absbipolar (cv):
          A variation of bipolar that always outputs a positive value, i.e. the pot will
          go 1 ... 0.5 ... 0 ... 0.5 ... 1
        lefthalf (cv):
          This output allows you to split the pot into two hemispheres. Here you get 1.0
          ... 0.0 while the pot is in the left half. In the middle and right of it you
          always get 0.
        righthalf (cv):
          This is the same but for the right half. It outputs 0 while the pot is in the
          left half and 0.0 ... 1.0 from the middle to the fully right position.
        lefthalfinv (cv):
          This outputs 1.0 - lefthalf, i.e. the value range 0.0 ... 1.0 ... 1.0 when the
          pot moves left → mid → right.
        righthalfinv (cv):
          This outputs 1.0 - righthalf, i.e. the value range 1.0 ... 1.0 ... 0.0 when
          the pot moves left → mid → right.
        comment: Add a comment in the droid ini file.

    """

    __ramsize__ = 40
    pot: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    notch: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    output: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    bipolar: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    absbipolar: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    lefthalf: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    righthalf: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    lefthalfinv: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    righthalfinv: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )



@dataclass
class Switchedpot(DroidCircuit):
    """Overlay pot with multiple functions (OBSOLETE).

    Args:
        pot (fraction):
          The pot that you want to overlay, e.g. P1.1
        bipolar (gate):
          If this input is set to 1, the usual pot range of 0 ... 1 will be mapped to -1
          ... +1, which converts this to a bipolar potentiometer. This is done by
          multiplying the output with 2.0 and substracting 1.0 afterwards.
        switch1 (gate):
          These inputs select which of the virtual pots should be changed when the
          physical pot is being turned. These should be set to 0 or 1 (or off and on).
        switch2 (gate):
          These inputs select which of the virtual pots should be changed when the
          physical pot is being turned. These should be set to 0 or 1 (or off and on).
        switch3 (gate):
          These inputs select which of the virtual pots should be changed when the
          physical pot is being turned. These should be set to 0 or 1 (or off and on).
        switch4 (gate):
          These inputs select which of the virtual pots should be changed when the
          physical pot is being turned. These should be set to 0 or 1 (or off and on).
        switch5 (gate):
          These inputs select which of the virtual pots should be changed when the
          physical pot is being turned. These should be set to 0 or 1 (or off and on).
        switch6 (gate):
          These inputs select which of the virtual pots should be changed when the
          physical pot is being turned. These should be set to 0 or 1 (or off and on).
        switch7 (gate):
          These inputs select which of the virtual pots should be changed when the
          physical pot is being turned. These should be set to 0 or 1 (or off and on).
        switch8 (gate):
          These inputs select which of the virtual pots should be changed when the
          physical pot is being turned. These should be set to 0 or 1 (or off and on).
        output1 (fraction):
          The output of the up to eight virtual pots.
        output2 (fraction):
          The output of the up to eight virtual pots.
        output3 (fraction):
          The output of the up to eight virtual pots.
        output4 (fraction):
          The output of the up to eight virtual pots.
        output5 (fraction):
          The output of the up to eight virtual pots.
        output6 (fraction):
          The output of the up to eight virtual pots.
        output7 (fraction):
          The output of the up to eight virtual pots.
        output8 (fraction):
          The output of the up to eight virtual pots.
        comment: Add a comment in the droid ini file.

    """

    __ramsize__ = 88
    pot: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    bipolar: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    switch1: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    switch2: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    switch3: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    switch4: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    switch5: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    switch6: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    switch7: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    switch8: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    output1: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    output2: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    output3: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    output4: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    output5: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    output6: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    output7: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )
    output8: Optional[str] = field(
            default=None,
            metadata=ctype.type_fraction(ramsize=0)
    )



@dataclass
class Togglebutton(DroidCircuit):
    """Create on/off buttons (OBSOLETE).

    Args:
        button (trigger):
          The actual push button. Usually you want to wire this to B1.1, B1.2 and so on:
          to one of the push buttons of your controllers. Each time that input goes from
          low to high the state of the push button will toggle.
        reset (trigger):
          A positive trigger edge here will reset the button into the state “not
          pressed” – regardless of its current state
        onvalue (cv):
          Value sent to output when the push button is on. Setting this to a different
          value than the default value saves you attenuating its value later on when you
          use it as a CV.
        offvalue (cv):
          Value sent to output when the push button is off.
        doubleclickmode (gate):
          This input can enable a double click mode when set to 1. In that mode the
          button only toggles it's constant state if you double press it in a short
          time. Otherwise it behaves like a momentary button, that inverts the persisted
          state (which you toggle with the double click).
        startvalue (gate):
          State of the push button when you switch on your system. Setting this to on or
          off will force the button into that state. Using this jack disables the
          persistence of the state! In switched mode this will be used for the other
          button layers as well.
        led (gate):
          When the button's state is on a value of 1.0 will be sent to that output –
          regardless of the values in onvalue and offvalue. Usually you will wire this
          jack to the LED within the button, e.g. to L1.1, L1.2 and so on
        output (cv):
          This jack will output either onvalue or offvalue depending on the state of the
          1 ... 4 button. If you have not wired those inputs then this is the same as
          the led output.
        inverted (cv):
          The same as output1, but sends onvalue when the button is off and offvalue
          when the button is on. Note: there is no inverted version of output2 ...
          output4.
        negated (gate):
          Similar to inverted, but always sends 1 when the button is off and 0 when the
          button is on – independent of the values of onvalue and offvalue.
        comment: Add a comment in the droid ini file.

    """

    __ramsize__ = 48
    button: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    reset: Optional[str] = field(
            default=None,
            metadata=ctype.type_trigger(ramsize=0)
    )
    onvalue: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    offvalue: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    doubleclickmode: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    startvalue: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    led: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )
    output: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    inverted: Optional[str] = field(
            default=None,
            metadata=ctype.type_cv(ramsize=0)
    )
    negated: Optional[str] = field(
            default=None,
            metadata=ctype.type_gate(ramsize=0)
    )






"""Circuit base class."""

from copy import copy
from dataclasses import dataclass, is_dataclass, fields, asdict
from typing import (
    ClassVar,
    Dict,
    Literal,
    List,
    Optional,
    Tuple,
    TypeVar,
)
from metapatch.base import Circuit


@dataclass
class DroidCircuit:
    """@private Droid Circuit parent class."""

    """@private"""
    comment: Optional[str] = None
    __ramsize__: ClassVar[int] = 0

    def to_circuit(self) -> Circuit:
        """Convert to circuit.

        @private
        """
        return dataclass_to_circuit(self)


T = TypeVar("T", bound=DroidCircuit)


def dataclass_to_circuit(
    circuit: DroidCircuit, section: Optional[str] = None, comment: Optional[str] = None
) -> Circuit:
    """Convert a dataclass to a circuit."""
    if not is_dataclass(circuit):
        raise ValueError("Error: Circuit must be a dataclass.")
    parameters: Dict[str, str] = {}
    for field in fields(circuit):
        value = getattr(circuit, field.name)
        field_name = field.name
        if not value:
            continue
        if field.name == "comment" and comment is None:
            comment = value
            continue
        if field.name.endswith("_"):
            field_name = field.name[:-1]
        parameters[field_name] = value

    circuit_name = type(circuit).__name__.lower()
    return Circuit(circuit_name, parameters, comment, section)


def transform(
    circuits: List[T],
    *,
    select: Optional[str] = None,
    select_at: Optional[str] = None,
    prepend: Optional[str] = None,
    append: Optional[str] = None,
    new_input: Optional[str] = None,
    new_output: Optional[str] = None,
    gate: Optional[str] = None,
    replace: Optional[List[Tuple[str, str]]] = None,
    ignore: Optional[List[str]] = None,
) -> List[T]:
    """Transform a list of circuits.

    Args:
        select: Add a select parameter to all circuits that support it and have
            no current select.
        select_at: Add a selectat parameter to all circuits that support it.
        prepend: Add a word to the start of all virtual cable names.
        append: Add a word to the end of all virtual cable names.
        new_input: If an input is found, change it to the value of input
        output: If an output is found, change it to value of output (e.g., O2)
        gate: If a gate is found, change it to the value of the gate.
        replace: List of (from, to). Does a search and replace for an arbitrary value.
           Can be used to e.g., replace one pot with another.
        ignore: Ignore any of the supplied names when doing a rewriting operation.

    """
    if not ignore:
        ignore = []
    if select:
        circuits = [add_select(circuit, select, select_at) for circuit in circuits]

    circuits = [rename_cables(circuit, append, prepend, ignore) for circuit in circuits]
    if new_input:
        circuits = change_jack(circuits, new_input, "input", ignore)
    if new_output:
        circuits = change_jack(circuits, new_output, "output", ignore)
    if gate:
        circuits = change_jack(circuits, gate, "gate", ignore)

    if replace:
        circuits = translate(circuits, replace)

    return circuits


def rename_cables(
    circuit: T, append: Optional[str], prepend: Optional[str], ignore: List[str]
) -> T:
    """Rename cables."""
    if not append and not prepend:
        return circuit

    cables = find_cables(circuit)
    if not cables:
        return circuit

    for fieldname, cablename in cables.items():
        if cablename in ignore:
            continue
        if not prepend:
            prependname = ""
        elif prepend.startswith("_"):
            prependname = prepend
        else:
            prependname = "_" + prepend

        newname = f"{prependname}{cablename}{append or ''}"
        setattr(circuit, fieldname, newname)

    return circuit


def find_cables(circuit: DroidCircuit) -> Dict[str, str]:
    """Return a dictionary of the input/outputs that contain virtual cables."""
    cables = {}
    for field in fields(circuit):
        value: str = getattr(circuit, field.name)
        if not value:
            continue
        if value.startswith("_"):
            cables[field.name] = value

    return cables


def add_select(circuit: T, select: str, select_at: Optional[str] = None) -> T:
    """Add select."""
    selectfield = [field for field in fields(circuit) if field.name == "select"]
    if not selectfield:
        return circuit
    select_val = getattr(circuit, "select")
    if not select_val:
        setattr(circuit, "select", select)

    if select_val != select:
        # Circuit has a different existing select value.
        return circuit

    if select_at:
        setattr(circuit, "selectat", select_at)

    return circuit


def change_jack(
    circuits: List[T],
    new_jack: str,
    jacktype: Literal["input", "output", "gate"],
    ignore: List[str],
) -> List[T]:
    """Change jack on a list of circuits.

    There can only be a single distinct value in all of the circuits.
    """
    jackmap = {"input": "I", "output": "O", "gate": "G"}
    start = jackmap[jacktype]
    jacks = set()
    new_circuits = []
    for circuit in circuits:
        circuit_fields = asdict(circuit)
        new_circuit = copy(circuit)
        for key, value in circuit_fields.items():
            if not value or value in ignore:
                continue
            if value.startswith(start):
                jacks.add(value)
                if len(jacks) > 1:
                    raise ValueError(
                        f"Cannot run {jacktype} transformation, "
                        f"more than one unique {jacktype} found."
                    )
                setattr(new_circuit, key, new_jack)
        new_circuits.append(new_circuit)

    return new_circuits


def translate(circuits: List[T], translations: List[Tuple[str, str]]) -> List[T]:
    """Translate parameters in multiple circuits.

    This allows renaming a controller to another controller.
    """
    new_circuits = []
    d_translation = dict(translations)
    for circuit in circuits:
        circuit_fields = asdict(circuit)
        c_copy = copy(circuit)
        for key, value in circuit_fields.items():
            if value and value in d_translation:
                setattr(c_copy, key, d_translation[value])
        new_circuits.append(c_copy)
    return new_circuits



"""Rudimentary droid patch parser."""

import re

from dataclasses import dataclass, field
from typing import Iterator, Dict, List, Optional


TCircuitParams = Dict[str, str]
TCircuit = Dict[str, TCircuitParams]
TSectionGroupedCircuits = Dict[str, List[TCircuit]]

TModDefinition = List[str]


@dataclass
class Token:
    """Token container class."""

    type: str
    value: str
    line: int
    column: int


@dataclass
class ParsedPatch:
    """Parsed patch container class."""

    sections: Dict[str, str] = field(default_factory=dict)
    circuits: List[Dict[str, Dict[str, str]]] = field(default_factory=list)
    controllers: List[str] = field(default_factory=list)

    @property
    def section_grouped(self) -> TSectionGroupedCircuits:
        """Group circuits by section."""
        current_section = "__none"
        grouped = {}
        for circuit in self.circuits:
            circuit_name = list(circuit.keys())[0]
            section = circuit[circuit_name].get("__section", current_section)
            if section not in grouped:
                grouped[section] = []
            grouped[section].append(circuit)
            current_section = section
        return grouped


def tokenize(patch: str) -> Iterator[Token]:
    """Tokenize a patch."""
    spec = [
        ("CONTROLLER", r"(?<=\[)([a-z][0-9]+[a-z0-9]*)(?=\])"),
        ("CIRCUIT", r"(?<=\[)([a-z]+)(?=\])"),
        ("PARAM", r"([a-z0-9]+)\s*(?= =)"),
        ("VALUE", r"(?<= =\s)([^\s].*)(?=\n)"),
        ("NEWLINE", r"\n"),
        ("COMMENT", r"(?<=#\s)(.*)(?=\n)"),
    ]

    tok_regex = "|".join("(?P<%s>%s)" % pair for pair in spec)
    line_num = 1
    line_start = 0
    section_re = re.compile(r"----*$")  # This is the regex that Forge uses too.
    in_section = False
    section_title_derived = False
    for mo in re.finditer(tok_regex, patch):
        # This currently throws away the patch title.
        # TODO: Handle top of patch comments
        kind = mo.lastgroup
        value = mo.group()
        column = mo.start() - line_start
        if kind == "NEWLINE":
            line_start = mo.end()
            line_num += 1
            continue
        if kind == "COMMENT" and section_re.match(value) is not None:
            in_section = not in_section
            if in_section:
                section_title_derived = False
            continue
        if in_section:
            if kind == "COMMENT":
                if section_title_derived:
                    kind = "SECTION_COMMENT"
                else:
                    kind = "SECTION"
                    section_title_derived = True

        if kind is None:
            continue
        yield Token(kind, value, line_num, column)


def parse_patch(patch: str) -> ParsedPatch:
    """Parse the patch."""
    sections = {}
    circuits = []
    controllers = []
    curr_params = {}
    curr_param: Optional[str] = None
    curr_section: Optional[str] = None
    curr_circuit: Optional[str] = None
    curr_circuit_comments: List[str] = []
    in_header = True
    for token in tokenize(patch):
        if token.type == "SECTION":
            curr_section = token.value
            sections[curr_section] = []
            in_header = False
            continue
        elif token.type == "CONTROLLER":
            # Discard all previous comments
            controllers.append(token.value)
            curr_circuit_comments = []

        elif token.type == "CIRCUIT":
            in_header = False
            if curr_circuit:
                circuits.append({curr_circuit: curr_params})
            curr_circuit = token.value
            curr_params = {}
            if curr_circuit_comments:
                curr_params["comment"] = "\\n".join(curr_circuit_comments)
                curr_circuit_comments = []
            if curr_section:
                curr_params["__section"] = curr_section

        elif token.type == "PARAM":
            curr_params[token.value] = ""
            curr_param = token.value
        elif token.type == "VALUE":
            curr_params[curr_param] = token.value

            curr_param = None
        elif token.type == "SECTION_COMMENT":
            sections[curr_section].append(token.value)
        elif token.type == "COMMENT":
            if not in_header:
                curr_circuit_comments.append(token.value)

    if curr_circuit:
        final_circuit = {curr_circuit: curr_params}
        if not circuits or final_circuit != circuits[-1]:
            circuits.append(final_circuit)

    return ParsedPatch(sections, circuits, controllers)



"""Class based patch generator interface."""

import argparse
import json
import textwrap

from abc import abstractmethod
from dataclasses import is_dataclass
from itertools import groupby
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
)

from .base import Circuit, Controller, Label
from .options import Option, BoolOption, NumberOption, EnumOption, Preset
from .utils import write_patch_section
from .circuits.base import dataclass_to_circuit, transform, DroidCircuit


# Some things to aid pdoc.
__docformat__ = "google"

DEFAULT_SECTION_NAME = "Generated Patch"


def _split_params(params: List[str]) -> Dict[str, str]:
    """Split input parameters."""
    patch_params: Dict[str, str] = {}
    for param in params:
        if "=" not in param:
            raise ValueError("Parameters must be input as key=value")
        key, value = param.split("=")
        if key in patch_params:
            raise ValueError(f"Duplicate key {key} found.")
        # We convert everything to strings.
        patch_params[key] = str(value)

    return patch_params


class MetaMetaPatch(type):
    """Metaclass for metapatches."""

    title: str
    description: str

    @property
    def synopsis(cls) -> Dict[str, Any]:
        """Get synopsis."""
        sections: Dict[str, Dict[str, Option]] = {}
        presets: List[Dict[str, Any]] = []
        for name, opt in cls.__dict__.items():
            if hasattr(opt, "_ispatchoption"):
                if opt.section not in sections:
                    sections[opt.section] = {}
                sections[opt.section][name] = opt

        presets = cls.presets
        section_list: List[Dict[str, Any]] = []
        for section_name, options in sections.items():
            subsection: Dict[str, Any] = {"title": section_name, "options": []}
            for option_name, option in options.items():
                subsection["options"].append(option.asdict(option_name))
            section_list.append(subsection)

        synopsis = {
            "title": cls.title,
            "description": cls.description,
            "sections": section_list,
            "presets": presets,
        }
        return synopsis

    @property
    def presets(cls) -> List[Dict[str, Any]]:
        """Fetch presets."""
        presets: List[Dict[str, Any]] = []
        for name, opt in cls.__dict__.items():
            if isinstance(opt, Preset):
                presets.append(opt.asdict(name))

        if presets:
            return presets
        # No presets defined.
        # We make a default preset.

        params = {}
        for name, opt in cls.__dict__.items():
            if hasattr(opt, "_ispatchoption"):
                params[name] = opt.default
        default_preset = {
            "name": "default",
            "title": "Default Preset",
            "parameters": params,
        }
        return [default_preset]


class PatchGenerator(metaclass=MetaMetaPatch):
    """Patch Generator.

    You need to subclass this to create your own patch generators.

    As a minimum, you need to provide the `title` and `description`
    class variables, and you should probably also provide one or more options.

    See `metapatch.option`.

    Presets can be defined with `metapatch.preset`.
    If no presets are defined, one named `default` will be created.

    For example:
    ```python
    class MyGenerator(metapatch.PatchGenerator)

        title = "My Patch Generator"
        description = "A simple patch generator"

        voices: int = metapatch.option("Number of voices", minimum=1, maximum=4)
        my_preset = metapatch.preset("Standard patch with 4 voices", {"voices": 4})
    ```

    """

    title: str
    description: str
    __patchvars__: List[str]

    def __new__(cls, **_: Any) -> "PatchGenerator":
        """Build class."""
        cls.__patchvars__ = []
        for param, opt in cls.__dict__.items():
            if hasattr(opt, "_ispatchoption"):
                cls.__patchvars__.append(param)

        return super().__new__(cls)

    def __init__(self, **kwargs: str) -> None:
        """Initialize class."""
        # Set all options not defined in the input to their defaults.
        for opt in self.__patchvars__:
            if opt not in kwargs:
                var = getattr(self, opt)
                assert issubclass(var, (BoolOption, EnumOption, NumberOption))
                setattr(self, opt, var.from_default().get_value())

        for key, value in kwargs.items():
            if key in self.__patchvars__:
                var = getattr(self, key)
                assert issubclass(var, (BoolOption, EnumOption, NumberOption))
                setattr(self, key, var(value).get_value())
            else:
                raise ValueError(f"Unknown parameter {key}")

        self._circuits: List[Circuit] = []
        self._controllers: List[Controller] = []
        self._labels: List[Label] = []
        self._section: Optional[str] = None
        # This dictionary stores comments for sections.
        self._sections: Dict[str, Optional[str]] = {}

    @classmethod
    def load_preset(cls, preset_name: str) -> "PatchGenerator":
        """Load preset.

        @private
        """
        preset = getattr(cls, preset_name)
        if preset:
            return cls(**preset.parameters)
        raise ValueError(f"Uknown preset: {preset_name}")

    @classmethod
    def _handle_cli(cls) -> argparse.Namespace:
        """Handle CLI."""
        parser = argparse.ArgumentParser(
            f'DROID patch generator "{cls.title}"',
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        parser.description = cls.description
        parser.add_argument(
            "parameters",
            nargs="*",
            help="Patch generator parameters.",
            metavar="param=value ...",
        )
        parser.add_argument(
            "--synopsis",
            "-s",
            action="store_true",
            help="Output possible parameters as JSON",
        )
        parser.add_argument(
            "-p", "--preset", metavar="P", help="Use settings from preset P"
        )
        presets = cls.presets
        presets_str = "\n".join(
            [f"{preset['name']:<12} {preset['title']:<12}" for preset in presets]
        )
        epilog = f"Available Presets:\n{presets_str}\n\n"
        params = "parameters (defaults are marked with *):\n\n"
        all_param_help = {}
        for key, value in cls.__dict__.items():
            if not hasattr(value, "_ispatchoption"):
                continue
            if issubclass(value, (BoolOption, EnumOption, NumberOption)):
                all_param_help[value.title] = value.help(key)

        if all_param_help:
            longest = max(
                [
                    len(varname[0])
                    for param in list(all_param_help.values())
                    for varname in param
                ]
            )
            for title, opthelp in all_param_help.items():
                params += textwrap.indent(title, "  ") + "\n"
                for opt, descr in opthelp:
                    params += textwrap.indent(f"{opt:<{longest}}  {descr}", "    ")
                    params += "\n"

                params += "\n"

            epilog += params
            parser.epilog = epilog

        args = parser.parse_args()

        return args

    @classmethod
    def run(cls) -> None:
        """Process parameters given via command line arguments.

        You need to call this function at the end of your script to make the
        patch generator work.

        ```python
        if __name__ == "__main__":
            MyPatchGenerator.run()
        ```

        """
        args = cls._handle_cli()
        if args.synopsis:
            print(json.dumps(cls.synopsis))
            return
        if args.preset:
            patch = cls.load_preset(args.preset)
            print(str(patch))
            return

        patch_parameters: List[str] = args.parameters
        parameters = _split_params(patch_parameters)

        patch = cls(**parameters)
        print(str(patch))

    def add(self, circuit: Any, comment: Optional[str] = None) -> None:
        """Add a defined circuit.

        This is used when adding one of the circuit classes.
        Args:
            circuit: The circuit class instance
            comment: Optional comment.
        """
        if not is_dataclass(circuit):
            raise ValueError("Unknown circuit type.")
        self._circuits.append(
            dataclass_to_circuit(circuit, self.section, comment=comment)
        )

    def add_circuit(
        self,
        name: str,
        params: Mapping[str, str],
        comment: Optional[str] = None,
    ) -> None:
        """Add a circuit.

        This allows you to generate raw DROID circuits, i.e., without using the
        circuit classes.

        Args:
            name: Circuit name, e.g. copy for a [copy] circuit
            params: Dictionary of circuit parameters.
            comment: Optional comment for the circuit.

        Example:
        ```python
        self.add_circuit(
            "button",
            {"button": "B1.1", "led": "L1.1", "output": "_MY_BUTTON"},
        )
        ```
        """
        self._circuits.append(
            Circuit(name=name, parameters=params, comment=comment, section=self.section)
        )

    def add_label(
        self, item: str, short_label: str, long_label: Optional[str] = None
    ) -> None:
        """Add a label.

        This enables you to add a label to a jack or a control.

        Args:
            item: string of the thing you want to label, e.g. O1 or G1.2
            short_label: The short label
            long_label: Optional longer label

        Example:
        ```python
        self.add_label("P1.1", "Activity", "Activity of algoquencer circuits.")
        ```
        """
        self._labels.append(Label.from_item(item, short_label, long_label))

    def _generate_labels(self) -> str:
        """Generate label strings."""
        sorted_labels: Dict[str, List[Label]] = {}
        for label in self._labels:
            if label.heading not in sorted_labels:
                sorted_labels[label.heading] = []
            sorted_labels[label.heading].append(label)

        labels = ""
        for heading, list_labels in sorted_labels.items():
            labels += f"# {heading.upper()}:\n"
            for label in list_labels:
                labels += str(label)
            labels += "\n"

        return labels

    def add_controller(self, type: str, position: int) -> None:
        """Add a controller at a given position.

        Args:
            type: Type of controller, e.g. B32
            position: controller position, e.g. 1
        """
        self._controllers.append(Controller(type, position))

    @property
    def section(self) -> Optional[str]:
        """Get the current section name, if any.

        This is implemented as a simple getter and setter.

        The getter holds the value of the current section.

        You can use the setter inline in your code to quickly change which
        section you are adding circuits to.

        While this will create a section if it does not exist, if you want to
        define your section properly, it's better to use
        `metapatch.PatchGenerator.add_section`.

        Example:
        ```python
        self.section = "Clock"
        self.add(metapatch.circuits.Lfo(hz="40", square="_CLOCK"))
        self.section = "Envelope"
        self.add(
            metapatch.circuits.Contour(
                trigger="_CLOCK",
                attack="0",
                decay="0.1",
                sustain="1",
                release="0.2",
                output="O1",
            )
        )
        ```

        """
        return self._section

    @section.setter
    def section(self, name: str) -> None:
        """Set the section."""
        if not isinstance(name, str):
            raise ValueError("Section name must be a string.")
        if name not in self._sections:
            self._sections[name] = None
        self._section = name

    def add_section(self, name: str, comment: Optional[str] = None) -> None:
        """Add a section with an optional comment.

        This will create a section in your patch.

        This differs from the `metapatch.DroidCircuit.section` setter in that it
        allows you to define a comment for your section.

        Args:
            name: Name of the section
            comment: Provide a comment for the section.

        """
        self._section = name
        if name not in self._sections or self._sections.get(name) != comment:
            self._sections[name] = comment

    def add_circuits(
        self,
        circuits: List[DroidCircuit],
        section: Optional[str] = None,
        *,
        select: Optional[str] = None,
        select_at: Optional[str] = None,
        prepend: Optional[str] = None,
        append: Optional[str] = None,
        input: Optional[str] = None,
        output: Optional[str] = None,
        gate: Optional[str] = None,
        replace: Optional[List[Tuple[str, str]]] = None,
        ignore: Optional[List[str]] = None,
    ) -> None:
        """Add multiple circuits.

        This allows you to add a collection of circuits, optionally into a named
        section, and optionally apply a set of transformations.

        Args:
            circuits: a list of Circuits.
            section: Name of the section to add them to.

        #### Transform actions:
        You can optionally add one or more of these transform actions:

        - **select**: Add a select parameter to any circuit that supports it.
        - **select_at**: Add or change the selectat attribute.
        - **prepend**: Prepend all virtual cable names with a string.
        - **append**: Append all virtual cable names with a string.
        - **output**: If an output is found, change it to value of input (e.g., O2)
        - **input**: If an input is found, change it to the value of input
        - **gate**: If a gate is found, change it to the value of the gate.
        - **replace**: List of (from, to). Does a search and replace for an given value.
                       Can be used to e.g., replace one pot with another.
        - **ignore**: Ignore any of the supplied names when doing a rewriting operation.

        #### Usage notes

        This method is intended to add a bunch of circuits all at once, for
        example if you have a function that generates some part of your patch
        logic.

        The transformation functions could for example allow you to generate a
        *generic* voice in a function, and then dynamically transform to scale
        to a number of voices.

        ```python

        # Generate the needed voices
        for voice in range(1, self.voices + 1):
            output_jack = "O" + str(voice)
            input_jack = "I" + str(voice)
            gate = "G1." + str(voice)
            self.add_circuits(
                self.generate_voice(),
                f"Voice {voice}",  # Use one section per voice
                select="_VOICE",  # Add a select
                select_at=str(voice), # Add a selectat matching the voice number
                input=input_jack,  # Rewrite inputs
                output=output_jack,  # Rewrite outputs
                gate=gate,  # Rewrite gates
                prepend=f"VOICE_{voice}",  # Prepend VOICE_X_ to all cables
                ignore=["_CLOCK", "_RESET"]  # Don't rewrite _CLOCK and _RESET
            )
        ```

        """
        assert all([is_dataclass(circuit) for circuit in circuits]) is True
        if not ignore:
            ignore = []

        action = {
            "select": select,
            "select_at": select_at,
            "prepend": prepend,
            "append": append,
            "new_input": input,
            "new_output": output,
            "gate": gate,
            "replace": replace,
        }
        # TODO: Test without the val is not None
        if any([val is not None for val in list(action.values())]):
            circuits = transform(circuits, ignore=ignore, **action)

        if not section:
            section = self.section
        for circuit in circuits:
            self._circuits.append(dataclass_to_circuit(circuit, section))

    @property
    def _has_sections(self) -> bool:
        """Check if sections are defined."""
        sections = [circuit.section for circuit in self._circuits]
        return any(sections)

    def _get_circuits_as_strings(self) -> str:
        """Get circuits as a string.

        This is for circuits when there are no sections at all.
        """
        circuits = [str(circuit) for circuit in self._circuits]
        return "\n".join(circuits)

    def _get_circuits_with_sections(self) -> str:
        """Get circuits as a string with sections."""
        circuits = []

        groups = {}
        for section, grouped_circuits in groupby(self._circuits, lambda x: x.section):
            if section is None:
                section_name = DEFAULT_SECTION_NAME
            else:
                section_name = section
            if section_name not in groups:
                groups[section_name] = []
            groups[section_name].extend(grouped_circuits)

        for section_name, grouped_circuits in groups.items():
            if section_name == DEFAULT_SECTION_NAME:
                circuits.append(write_patch_section(DEFAULT_SECTION_NAME))
            else:
                comment = self._sections.get(section_name)
                circuits.append(write_patch_section(section_name, comment))

            for circuit in grouped_circuits:
                circuits.append(str(circuit))

        return "\n".join(circuits)

    def __str__(self) -> str:
        """Output patch as string."""
        self._generate()

        labels = self._generate_labels()
        sorted_controllers = sorted(self._controllers, key=lambda x: x.position)
        controllers = "\n".join([str(controller) for controller in sorted_controllers])
        if self._has_sections:
            circuits = self._get_circuits_with_sections()
        else:
            circuits = self._get_circuits_as_strings()

        return f"{labels}{controllers}\n\n{circuits}"

    def _generate(self) -> None:
        """Generate patch.

        This is an internal function that ensures that we reset the patch if needed.
        """
        if not self._circuits and not self._controllers:
            self.generate()

    @abstractmethod
    def generate(self) -> None:
        """Generate patch.

        Define this in your subclass, and use it to add circuits and controllers.

        Example:
        ```python
        def generate(self) -> None:
            # Add some controllers
            self.add_controller("p4b2", 1)
            self.add_controller("b32", 2)
            # Add a LFO
            self.add(metapatch.circuits.Lfo(hz="-8 * P1.1 + 8"), square="_LFO")
            # ...

        ```
        """



"""DROID circuits JSON parsing."""

import json
import keyword
from pathlib import Path

from typing import Dict, List, TypedDict

from .types import Circuit, CircuitCollection, Parameter


class DroidControllerJSON(TypedDict):
    """Typing of the controllers JSON."""

    ramsize: int


class DroidParamJSON(TypedDict):
    """Typing of the DROID circuit JSON."""

    type: str
    ramsize: int
    count: int
    default: str
    prefix: str
    name: str
    start_at: int
    description: str


class DroidCircuitJSON(TypedDict):
    """Typing of the DROID circuit JSON."""

    category: str
    ramsize: int
    title: str
    description: str
    inputs: List[DroidParamJSON]
    outputs: List[DroidParamJSON]


class DroidOuterJSON(TypedDict):
    """Outer JSON."""

    firmware_version: str
    circuits: Dict[str, DroidCircuitJSON]
    controllers: Dict[str, DroidControllerJSON]
    available_memory: Dict[str, int]


def _parse_parameters(param: DroidParamJSON, direction: str) -> List[Parameter]:
    """Parse parameters."""
    params = []
    if "prefix" in param:
        prefix = param["prefix"]
        start = param["start_at"]
        count = param["count"]
        for step in range(start, count + start):
            p_name = prefix + str(step)
            params.append(
                Parameter(
                    p_name,
                    param["type"],
                    param.get("description", ""),
                    param.get("ramsize", 0),
                    direction,
                )
            )
    else:
        p_name = param["name"]
        if keyword.iskeyword(p_name):
            p_name = f"{p_name}_"
        params.append(
            Parameter(
                p_name,
                param["type"],
                param.get("description", ""),
                int(param.get("ramsize", 0)),
                direction,
            )
        )

    return params


def _parse_circuits(json_circuits: DroidOuterJSON) -> List[Circuit]:
    """Perform the actual parsing."""
    circuits: List[Circuit] = []

    for circuit_name, params in json_circuits["circuits"].items():
        inputs = params.get("inputs", [])
        outputs = params.get("outputs", [])
        title = params.get("title", "").strip()
        category = params.get("category", "other")
        ramsize = params.get("ramsize", 0)
        if not ramsize:
            ramsize = 0
        circuit = Circuit(circuit_name, title, category, ramsize=ramsize)
        for in_param in inputs:
            circuit.parameters.extend(_parse_parameters(in_param, "input"))

        for out_param in outputs:
            circuit.parameters.extend(_parse_parameters(out_param, "output"))

        circuits.append(circuit)

    return circuits


def load_circuits(filename: Path) -> CircuitCollection:
    """Load circuits from file."""
    assert filename.exists(), "Error: Filename does not exist!"
    with open(filename, "r", encoding="utf-8") as f:
        raw_circuits: DroidOuterJSON = json.load(f)

    circuits = _parse_circuits(raw_circuits)
    droid_version = raw_circuits["firmware_version"]

    return CircuitCollection(droid_version, circuits)
